const Promise = require('bluebird');
const _ = require('lodash');
const logger = require('cf-logs').Logger('codefresh:variable-resolver');
const { split, withTimeout, to } = require('./utils');
const Vars = require('./Vars');
const errors = require('./errors');

/** @module @codefresh-io/cf-variable-resolver */

/**
 * Logger
 * @typedef {Object} Logger
 * @property {function(string):undefined} info - log info level
 * @property {function(string):undefined} warn - log warn level
 */

/**
 * getContextFn
 * @typedef {function(string, Object, Logger): Promise<*>} getContextFn
 */

/**
 * Resolver
 * @typedef {Object} Resolver
 * @property {function(string, getContextFn, Logger):Promise<string>} resolve - resolution function, gets: variable, getContextFn, and logger
 */

/**
 * Resolution
 * @typedef {Object} Resolution
 * @property {number} index - index
 * @property {string} key - key
 * @property {boolean} resolved - resolved
 * @property {string} value - value
 */

/**
 * PathResolution
 * @typedef {Object} PathResolution
 * @property {string} path - path
 * @property {Resolution} resolution - resolution
 */

/**
 * Variable resolver options
 * @typedef {Object} Options
 * @property {number | undefined} singleResolutionTimeout - sets the timeout for single variable resolution
 * @property {number | undefined} completeResolutionFlowTimeout - sets the timeout for the entire resolution
 * @property {Logger | undefined} log - sets the timeout for the entire resolution
 * @property {undefined | function(string, Object, Logger): Promise<*>} getContext - a function used by the variable resolver to retrieve contexts
 * @property {undefined | function(Resolution, Logger): Promise<void>} addNewMask - a function used by the variable resolver to add a new mask
 */

/**
 * resolution options
 * @typedef {Object} ResolutionOptions
 * @property {number | undefined} timeout - sets the timeout for single variable resolution
 * @property {Logger | undefined} log - sets the timeout for the entire resolution
 * @property {function(string, Object, Logger): Promise<*>} getContext - a function used by the variable resolver to retrieve contexts
 * @property {function(Resolution, Logger): Promise<void>} addNewMask - a function used by the variable resolver to add a new mask
 */

/**
 * VariableResolver
 * @typedef {Object} VariableResolver
 * @property {function(string, ResolutionOptions): Promise<Resolution[]>} resolve - resolve a string variable
 * @property {function(Object, ResolutionOptions): Promise<PathResolution[]>} recursive - recursively resolve secrets on an object
 * @property {function(Object, ResolutionOptions): Promise<Object>} resolveSecretReferences - recursively resolves secrets on a context according to secretStoreReferences field
 * @property {function(getContextFn)} setGetContext - sets the getContext function of the variable resolver
 * @property {function(function(Resolution, Logger): Promise<void>)} setAddNewMask - sets the addNewMask function of the variable resolver
 */

const ignoreKeys = [
    'originalCommands',
];

/**
 * Builds a new variable resolver
 * @param {Object.<string, Resolver>} resolvers resolvers
 * @param {Options} options options
 * @returns {VariableResolver} variableResolver
 */
function build(resolvers, options) {
    const resolutionTimeout = _.get(options, 'completeResolutionFlowTimeout');
    const self = { resolvers, ...{ log: logger }, ...options };
    const resolve = _resolveString.bind(self);
    const recursive = _recursive.bind(self);
    const resolveSecretReferences = _resolveSecretReferences.bind(self);

    return {
        resolve: resolutionTimeout ? withTimeout(resolve, resolutionTimeout) : resolve,
        recursive: resolutionTimeout ? withTimeout(recursive, resolutionTimeout) : recursive,
        resolveSecretReferences: resolutionTimeout ? withTimeout(resolveSecretReferences, resolutionTimeout) : resolveSecretReferences,
        setGetContext: (getContext) => { self.getContext = getContext; },
        setAddNewMask: (addNewMask) => { self.addNewMask = addNewMask; },
    };
}

/**
 * resolve a string that contains variables
 * @this {Options & { resolvers: Object.<string, Resolver> }}
 * @param {string} variable variable to resolve
 * @param {ResolutionOptions} options options
 * @returns {Promise<Resolution[]>} resolutions
 */
async function _resolveString(variable, options) {
    const vars = new Vars(variable);
    if (!vars.valueContainsVariables()) {
        return [{
            index: 0,
            key: variable,
            value: variable || '',
            resolved: false,
        }];
    }

    const { resolvers } = this;
    const keys = vars.getVariableKeys();
    if (!resolvers) {
        // resolver not found
        // return the variable as-is
        return [{
            index: 0,
            key: variable,
            value: variable,
            resolved: false,
        }];
    }

    return Promise.mapSeries(keys, async (key, index) => _resolveKey({
        resolvers,
        key,
        index,
        timeout: _.get(options, 'timeout', this.singleResolutionTimeout),
        log: _.get(options, 'log', this.log),
        getContext: _.get(options, 'getContext', this.getContext),
        addNewMask: _.get(options, 'addNewMask', this.addNewMask),
    }));
}

/**
 * recurse through an object and resolve all the variables it contains
 * @this {Options & { resolvers: Object.<string, Resolver> }}
 * @param {*} obj object with variables to resolve
 * @param {ResolutionOptions} options options
 * @returns {Promise<PathResolution[]>} resolutions
 */
async function _recursive(obj, options) {
    const result = [];
    if (!obj) {
        return result;
    }
    if (_.isString(obj)) {
        result.push({
            path: '',
            resolution: await _resolveString.bind(this)(obj, options),
        });
    } else if (_.isArray(obj)) {
        await Promise.mapSeries(obj, async (item, index) => {
            const resolution = await _recursive.bind(this)(item, options);
            _.map(resolution, (r) => {
                let path = `[${index}]`;
                if (r.path) {
                    path = `${path}${r.path}`;
                }
                result.push({
                    path,
                    resolution: r.resolution,
                });
            });
        });
    } else if (_.isObject(obj)) {
        await Promise.mapSeries(Object.keys(_.omit(obj, ignoreKeys)), async (key) => {
            const resolution = await _recursive.bind(this)(obj[key], options);
            _.map(resolution, (r) => {
                let path = `[${key}]`;
                if (r.path) {
                    path = `${path}${r.path}`;
                }
                result.push({
                    path,
                    resolution: r.resolution,
                });
            });
        });
    }

    return result;
}

/**
 * resolveKey options
 * @typedef {Object} ResolveKeyOptions
 * @property {Object.<string, Resolver>} resolvers - resolvers
 * @property {string} key - key
 * @property {number} index - index
 * @property {?number} timeout - timeout for single variable resolution
 * @property {Logger} log - logger
 * @property {function(string, Object, Logger): Promise<*>} getContext - a function used by the variable resolver to retrieve contexts
 * @property {function(Resolution, Logger): Promise<void> | undefined} addNewMask - a function used by the variable resolver to add a new mask
 */

/**
 * resolves a single variable
 * @param {ResolveKeyOptions} options - options
 * @returns {Promise<Resolution>} - resolution
 */
async function _resolveKey({
    resolvers, key, index, timeout, log, addNewMask, getContext
}) {
    const parts = split(key);
    const resolver = _.head(parts);
    if (!resolvers[resolver]) {
        return {
            index,
            key: `$\{{${key}}}`,
            value: `$\{{${key}}}`,
            resolved: false,
        };
    }
    const res = {
        index,
        key: `$\{{${key}}}`,
    };
    log.info(`Resolving variable $\{{${key}}}`);
    let resolutionFn = resolvers[resolver].resolve;
    if (timeout) {
        resolutionFn = withTimeout(resolutionFn, timeout);
    }
    const [err, value] = await to(async () => resolutionFn(`$\{{${key}}}`, getContext, log));
    if (err) {
        log.warn(`Failed to resolve variable: $\{{${key}}}. ${err.message}`);
        logger.error(err.stack);
        if (_.has(err, 'suggestion')) {
            log.warn(err.suggestion);
        }
        if (_.get(err, 'lvl', errors.ErrorLVL.Warn) === errors.ErrorLVL.Crit) {
            throw err;
        }
        const istimeout = _.get(err, 'name', '') === 'TimeoutError';
        if (_.has(err, 'default') || istimeout) {
            res.value = err.default || '';
            res.resolved = true;
        } else {
            res.value = `$\{{${key}}}`;
            res.resolved = false;
        }

    } else {
        res.value = value;
        res.resolved = true;

        if (resolvers[resolver].secret && addNewMask) {
            log.info(`Creating logs mask for variable $\{{${key}}}`);

            const [_err] = await to(addNewMask(res, log));
            if (_err) {
                log.warn(`Failed to create mask for variable: ${res.key}`);
            }
        }
    }
    return res;
}

/**
 * recursively resolves secrets on a context according to secretStoreReferences field
 * @this {Options & { resolvers: Object.<string, Resolver> }}
 * @param {Object} context context with secret references to resolve
 * @param {ResolutionOptions} options options
 * @returns {Promise<Object>} resolutions
 */
async function _resolveSecretReferences(context, options) {
    const log = _.get(options, 'log', this.log);
    const contextName = _.get(context, 'metadata.name');

    const secretStoreReferences = _.get(context, 'spec.data.secretStoreReferences', []);
    const secretResolvePromises = secretStoreReferences.map(async (secretPropPath) => {
        const fullSecretPropPath = `spec.data.${secretPropPath}`;
        const secretRef = _.get(context, fullSecretPropPath);
        if (!secretRef) {
            log.warn(`ignoring reference '${secretPropPath}' in context '${contextName}' since it does not exist on the context`);
            return;
        }
        try {
            const res = await _resolveString.call(this, secretRef, options);
            const [{ resolved, value }] = res;
            if (!resolved) {
                throw new Error(`failed to resolve secret-store reference '${secretRef}' in context '${contextName}'`);
            }

            _.set(context, fullSecretPropPath, value);
        } catch (err) {
            throw new Error(`failed to resolve secret-store reference '${secretRef}' in context '${contextName}' caused by: ${err}`);
        }
    });

    await Promise.all(secretResolvePromises);
    _.unset(context, 'spec.data.secretStoreReferences');
    return context;
}

exports.build = build;
